from .define import define_log_model  # noqa
from .models import BaseLogEntry
from .util import LogEntryKindMap

LogEntryKind = LogEntryKindMap()

__all__ = ("define_log_model", "LogEntryKind", "BaseLogEntry")



from django.db import models

from .models import BaseLogEntry

all_known_log_models = {}


def define_log_model(
    model_class,
    base_class=BaseLogEntry,
    on_delete=models.CASCADE,
    allow_null_target=False,
):
    """
    Define a log model for a given Django model class ("parent model").

    The log entry model class is returned, and it must be assigned to
    a variable in a `models.py` file to allow Django to pick it up
    for migrations.

    For all intents and purposes, the log entry model is owned by your
    app after this function creates it.

    The parent model is anointed with two new attributes:

    * ``add_log_entry``, a function to add a log entry. (See
      :func:`BaseLogEntry.add_log_entry` for details,
      but heed the fact that the ``target`` argument will be
      implicitly passed.)
    * ``log_model``, a reference to the log entry model.

    :param model_class: The model class this log entry model is for.
    :param base_class: Replacement base class for the model. Should
                       be compatible with :class:`BaseLogEntry` never-
                       theless.
    :param on_delete: The `on_delete` clause for the log entry class's
                      foreign key. Defaults to `CASCADE`, i.e. that
                      log entries are deleted when the logged model
                      instance is.  `PROTECT` would be another sane
                      option.
    :param allow_null_target: Whether to allow null values as target.
                              This lets you have "generic" model log
                              entries in the same table as instance
                              specific ones.
    :type allow_null_target: bool
    :return: The log entry model.
    """
    log_model_name = "%sLogEntry" % model_class.__name__
    allow_null_target = bool(allow_null_target)

    class Meta:
        app_label = model_class._meta.app_label
        abstract = False

    class_dict = {
        "target": models.ForeignKey(
            to=model_class,
            related_name="log_entries",
            on_delete=on_delete,
            blank=allow_null_target,
            null=allow_null_target,
        ),
        "__module__": model_class.__module__,
        "Meta": Meta,
        "logged_model": model_class,
    }

    log_entry_class = type(str(log_model_name), (base_class,), class_dict)

    def add_log_entry(
        self, message, identifier=None, kind="other",
        user=None, extra=None, save=True, **kwargs
    ):
        return log_entry_class.add_log_entry(
            target=self,
            message=message,
            identifier=identifier,
            kind=kind,
            user=user,
            extra=extra,
            save=save,
            **kwargs
        )

    setattr(model_class, "add_log_entry", add_log_entry)
    setattr(model_class, "log_model", log_entry_class)
    all_known_log_models[model_class] = log_entry_class
    return log_entry_class



from analog.exceptions import UnknownLogKind
from analog.settings import KINDS


class LogEntryKindMap:
    """
    A helper class for transitioning old code.

    Allows looking up log entry kinds by "enumish" name, i.e.
    ``LogEntryKind.OTHER`` would map to the "other" kind's ID.
    """

    def __getattr__(self, kind):
        """
        Get the kind ID for the given kind mnemonic.

        :param kind: Kind mnemonic.
        :type kind: str
        :return: int
        """
        if kind.startswith('_'):
            return self.__getattribute__(kind)
        kind = str(kind).lower()
        if kind not in KINDS:
            raise UnknownLogKind(kind)
        return KINDS[kind]



class UnknownLogKind(ValueError):
    """Exception thrown when an unknown ``kind`` is passed."""

    def __init__(self, value):
        """
        Construct the exception.

        :param value: The invalid kind value passed in.
        """
        message = "Unknown log entry kind %r" % value
        super().__init__(message)


class NoExtraField(ValueError):
    pass



from django.conf import settings
from django.db import models
from django.utils.encoding import force_str

from analog.exceptions import NoExtraField, UnknownLogKind
from analog.settings import KIND_IDS, KIND_LABELS, KINDS


def _map_kind(kind):
    if isinstance(kind, str):
        if kind not in KINDS:
            raise UnknownLogKind(kind)
        kind = KINDS[kind]
    assert isinstance(kind, int)
    if kind not in KIND_IDS:
        raise UnknownLogKind(kind)
    return kind


class BaseLogEntry(models.Model):
    """
    Abstract base model class for the various log models.

    The concrete models are created by :func:`define_log_model`.

    In addition, directly deriving a model from `BaseLogEntry` is supported
    (for instance, to allow for log entries that are not attached to other
    models), though the `add_log_entry` function will naturally not be
    automatically augmented to pass a `target` parameter.
    """

    target = None  # This will be overridden dynamically
    created_on = models.DateTimeField(auto_now_add=True, editable=False)
    user = models.ForeignKey(
        getattr(settings, "AUTH_USER_MODEL", "auth.User"),
        null=True, on_delete=models.PROTECT
    )
    message = models.CharField(max_length=256)
    identifier = models.CharField(max_length=64, blank=True)
    kind = models.IntegerField(default=0, db_index=True)

    class Meta:
        abstract = True

    def get_kind_display(self):
        """
        Get the kind label for this log entry.

        This emulates the behavior Django would have for fields that
        have the ``choices`` kwarg.
        """
        return KIND_LABELS.get(self.kind, self.kind)

    def save(self, *args, **kwargs):
        if self.pk:
            raise ValueError("%r objects may not be modified" % self.__class__)
        super().save(*args, **kwargs)

    @classmethod
    def add_log_entry(
        cls,
        target, message, identifier=None, kind="other",
        user=None, extra=None, save=True,
        **kwargs
    ):
        """
        Add a log entry.

        .. note::

           This method should not be used directly; instead, the same method,
           aside from the `target` argument, is available on models that
           have been anointed by :func:`define_log_model`.

        :param target: Target model instance
        :param message: Log message
        :type message: str
        :param identifier: Log message identifier.
                           Useful for tagging log entries in a way that is
                           not necessarily human-readable.
        :type identifier: str|None
        :param kind: Log entry kind. Either a mnemonic string (preferred and
                     readable), or the actual entry ID integer (if you really
                     have to). Kinds are configured in your project's settings.
        :type kind: int|str
        :param user: An optional user object (an instance of
                     ``settings.AUTH_USER_MODEL``) to attach to this log entry.
        :param extra: Extra data, if applicable. If set, this must be
                      serializable to JSON; ``dict``s are a good idea.
                      This is only applicable if the model contains an `extra`
                      field; passing non-None here if the model
                      does not contain `extra` is an error.
        :type extra: object|None
        :param kwargs: Any other fields to pass to the constructor of the
                       class. Mainly useful with log classes derived from
                       `BaseLogEntry`.
        :type kwargs: dict
        :param save: Whether to immediately save the log entry. Default True.
        :type save: bool
        :return: The created log entry
        """

        if target is not None and not getattr(target, "pk", None):
            raise ValueError("Can not create log entry for unsaved object")

        kind = _map_kind(kind)

        if not getattr(user, "pk", None):
            user = None

        kwargs = dict(
            target=target,
            message=message,
            identifier=force_str(identifier or "", errors="ignore")[:64],
            user=user,
            kind=kind,
            **kwargs
        )

        has_extra_field = ('extra' in cls._meta._forward_fields_map)
        if extra is not None:
            if not has_extra_field:
                raise NoExtraField(
                    'The %r class has no `extra` field,'
                    'but non-None extra was passed!' % cls
                )
            kwargs['extra'] = extra

        if target is None:
            kwargs.pop('target')
        log_entry = cls(**kwargs)
        log_entry.clean()
        if save:
            log_entry.save()
        return log_entry



from django.conf import settings
from django.utils.translation import gettext_lazy as _

_DEFAULT_KINDS = {
    "other": 0,
    "audit": 1,
    "edit": 2,
    "deletion": 3,
    "note": 4,
    "email": 5,
    "warning": 6,
    "error": 7
}

_DEFAULT_KIND_LABELS = {
    "other": _("other"),
    "audit": _("audit"),
    "edit": _("edit"),
    "deletion": _("deletion"),
    "note": _("note"),
    "email": _("email"),
    "warning": _("warning"),
    "error": _("error")
}


def _get_kind_labels(kinds):
    kind_labels = _DEFAULT_KIND_LABELS.copy()
    kind_labels.update(getattr(settings, "ANALOG_KIND_LABELS", {}))
    kind_labels = {
        KINDS.get(mnemonic, mnemonic): label
        for (mnemonic, label) in kind_labels.items()
    }
    for key in list(kind_labels):
        if isinstance(kind_labels[key], str):
            kind_labels[key] = _(kind_labels[key])
    return kind_labels


#: A mapping from kind mnemonic (string) to kind ID (integer)
KINDS = _DEFAULT_KINDS.copy()
KINDS.update(getattr(settings, "ANALOG_KINDS", {}))

#: A mapping from kind ID (integer) to label
KIND_LABELS = _get_kind_labels(KINDS)

KIND_IDS = set(KINDS.values())
